/*
	getelementsbyselector.js
	Copyright (C) 2005 H. Adam Mill II

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

		 http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	Reference:
		http://www.w3.org/TR/CSS21/selector.html
		
	History:
		2005-08-26	: 	Added additional test cases.
								Added grouping support. (ex. .blue, .white)
								Added attribute "|=" (starts-with, hyphen-separated) support. (ex. div[lang|="en"])
								Improve commenting a bit.
		2005-08-14	:	Added support for sibling selectors (ex. div + p).
								Fix issue with attribute selector values containing "]" (close bracket) failing. (ex. a[href="javascript:alert('[ some text ]');"])
								Improve error handling.
								Spelling fixes, formatting changes, and minor optimizations.
								Removed unused parts of supporting utils.js file.
		2005-08-09	:	Initial release.

	TODOs:
		- ":first-child" support.
		- Namespace support.
		- Build a compiled selector cache.
		- Fix anything in the code that says TODO.
		- Test in OS X & Linux specific browsers.

	Known Issues:
		- IE takes significantly longer to execute the tests than Gecko or Opera. Need to profile and fix if possible.

	Tested In:
		- IE 6
		- Opera 8; Opera 7.54
		- Firefox 1.04, 1.06;  Firefox Deer Park Alpha 2
		- Mozilla 1.7.8; Mozilla 1.7.11

	This script depends on utils.js (included in the distribution).
*/


// Add getElementsBySelector to the document instance.
document.getElementsBySelector = function(selector)
{
	return DOMUtils.getElementsBySelector(document, selector);
}


extend(DOMUtils, {
	// Get a list of elements matching a CSS selector.
	getElementsBySelector : function(parent, selector)
	{
		var compiled = DOMUtils.getCompiledSelector(selector);
		return DOMUtils.getElementsByCompiledSelector(parent, compiled);
	},
	// Get a list of elements matching compiled selector.
	getElementsByCompiledSelector : function(parent, compiled, depth)
	{
		if(!compiled || !compiled.length)
			return null;
		if(!parent) parent = document;
		var elementlist = [ ];		
		if(undefined == depth)
		{
			for(var loop=0; loop < compiled.length; loop++)
				elementlist = elementlist.concat(DOMUtils.getElementsByCompiledSelector(parent, compiled[loop], 0));
			elementlist = ArrayUtils.removeDuplicates(elementlist);
		}
		else
		{
			var selector = compiled[depth];
			if(parent.parentNode)
				var grandparent = parent.parentNode;
			else
				var grandparent = null;
			switch(selector.type)
			{
				case "id":
				{
					var element = null;
					if("all" == selector.descend)
						element = DOMUtils.getElementById(parent, selector.data);
					else if("child" == selector.descend)
						element = DOMUtils.getChildById(parent, selector.data);
					else if("self" == selector.descend)
					{
						if(parent.id && parent.id == selector.data)
							element = parent;
					}
					else if("sibling" == selector.descend && grandparent)
					{
						element = DOMUtils.getChildById(grandparent, selector.data);
						if(element && parent != DOMUtils.getPreviousSiblingElement(element))
							element = null;
					}
					if(element)
						elementlist = [ element ];							
					break; 
				}
				case "class":
				{
					if("all" == selector.descend)
						elementlist = DOMUtils.getElementsByClassName(parent, selector.data);											
					else if("child" == selector.descend)
						elementlist = DOMUtils.getChildrenByClassName(parent, selector.data);
					else if("self" == selector.descend)
					{
						if(parent.className)
						{
							var re_classname = RegExpUtils.getWhiteSpaceListMatcher(selector.data);
							if(re_classname.test(parent.className))
								elementlist = [ parent ];
						}
					}
					else if("sibling" == selector.descend && grandparent)
					{
						siblinglist = DOMUtils.getChildrenByClassName(grandparent, selector.data);
						elementlist = DOMUtils._getSelectorSiblingMatch(parent, siblinglist);
					}
					break;
				}
				case "tag":
				{
					if("all" == selector.descend)
						elementlist = ArrayUtils.toArray(parent.getElementsByTagName(selector.data));
					else if("child" == selector.descend)
						elementlist = ArrayUtils.toArray(DOMUtils.getChildrenByTagName(parent, selector.data));
					else if("self" == selector.descend)
					{
						if(parent.nodeName && parent.nodeName == selector.data)
								elementlist = [ parent ];				
					}
					else if("sibling" == selector.descend && grandparent)
					{
						siblinglist = DOMUtils.getChildrenByTagName(grandparent, selector.data);
						elementlist = DOMUtils._getSelectorSiblingMatch(parent, siblinglist);
					}
					break;
				}
				case "*":
				{
					if("all" == selector.descend)
						elementlist = ArrayUtils.toArray(DOMUtils.getAllChildren(parent));
					else if("child" == selector.descend)
						elementlist = ArrayUtils.toArray(DOMUtils.getDirectChildren(parent));
					else if("sibling" == selector.descend && grandparent)
					{
						siblinglist = ArrayUtils.toArray(DOMUtils.getAllChildren(grandparent));
						elementlist = DOMUtils._getSelectorSiblingMatch(parent, siblinglist);
					}
					break;
				}
				case "attr":
				{
					var element = null;
					var attrib = parent.getAttribute(selector.data.name);
					if(attrib)
					{
						switch(selector.data.cmp)
						{
							case "*":
							{
								element = parent;
								break;
							}
							case "starts":
							{
								if(selector.data.value == attrib.substr(0, selector.data.value.length))
									element = parent;
								break;
							}
							case "eq":
							{
								if(selector.data.value == attrib)
									element = parent;
								break;
							}
							case "ends":
							{
								if(selector.data.value == attrib.substr(attrib.length-selector.data.value.length))
									element = parent;
								break;
							}						
							case "contains":
							{
								if(-1 != attrib.indexOf(selector.data.value))
									element = parent;
								break;
							}
							case "one-of":
							{										
								var re_oneof = RegExpUtils.getWhiteSpaceListMatcher(selector.data.value);
								if(re_oneof.test(attrib))
									element = parent;
								break;
							}
							case "hyphen":
							{
								var re_hypen = new RegExp("^(" + RegExpUtils.escape(selector.data.value) + ")(-|$)", "");
								if(re_hypen.test(attrib))
									element = parent;
								break;
							}
						}
						if(element)
							elementlist = [ element ];
					}
					break;
				}
				case "psudo":
				{
					if("root" == selector.data && 0 == depth)
					{
						if(document.documentElement)
							elementlist = [ document.documentElement ];
						else
							elementlist = [ document ];
					}
					break;
				}
			}
			if(elementlist && elementlist.length && depth+1 < compiled.length)
			{
				var parentlist = elementlist;
				var elementlist = [ ];
				for(var loop=0; loop < parentlist.length; loop++)
				{
					var eles = DOMUtils.getElementsByCompiledSelector(parentlist[loop], compiled, depth+1);
					if(null != eles && eles.length)
						elementlist = elementlist.concat(eles);
				}
			}
		}
		return elementlist;
	},
	// Compile a CSS selector for use with getElementsByCompiledSelector.
	getCompiledSelector : function(selector)
	{
		var compiledlist = [ ];
		var compiledselector = [ ];
		var nextselector = false;
		var partlist = StringUtils.breakup(StringUtils.trim(selector), " \t.#+>:[],");
		var descend = "all";
		// Compiled object is has the following members: descend, type, data, source.
		var compiled = null;
		for(var loop=0; loop < partlist.length; loop++)
		{
			var part = partlist[loop];
			var source = part;
			switch(part)
			{
				case "+":
				{
					descend = "sibling";
					break;
				}
				case ">":
				{
					descend = "child";
					break;
				}
				case "\t":
				case " ":
				{
					if("self" == descend)
						descend = "all";
					break;
				}
				case ":":
				case "#":
				case ".":
				{
					if(loop+1 >= partlist.length)
						return null;
					if("#" == part)
						type = "id";
					else if(":" == part)
						type = "psudo";
					else
						type = "class";
					loop++;
					part = partlist[loop];
					source += part;
					compiled = { descend : descend, type : type, data : part };
					break;
				}
				case "[":
				{
					var attrib = { };
					source = "";
					loop++;
					// Recombine parts that were broken up until the entire attribute selector is found.
					while(loop < partlist.length)
					{
						part = partlist[loop];
						if("]" == part && !(StringUtils.count(source, "\"")%2)) // TODO: This is ugly, replace it.
							break;
						source += part;
						loop++;
					}
					if("]" != part)
						return null;
					var match = DOMUtils._re_attrib.exec(source);
					if(!match)
						return null;
					attrib.name = match[1];
					if('' == match[2])
					{
						attrib.cmp = "*";
						attrib.value = null;
					}
					else
					{
						switch(match[3])
						{
							case "=":
							{
								attrib.cmp = "eq";
								break;
							}
							case "^=":
							{
								attrib.cmp = "starts";
								break;
							}
							case "$=":
							{
								attrib.cmp = "ends";
								break;
							}
							case "*=":
							{
								attrib.cmp = "contains";
								break;
							}						
							case "~=":
							{
								attrib.cmp = "one-of";
								break;
							}
							case "|=":
							{
								attrib.cmp = "hyphen";
								break;
							}
							default:
								return null;
						}
						attrib.value = match[4];
					}
					source = "[" + source + "]";
					compiled = { descend : "self", type : "attr", data : attrib };
					break;
				}
				case "*":
				{
					compiled = { descend : descend, type : "*", data : part };
					break;
				}
				case ",":
				{
					nextselector = true;
					break;
				}
				default:
				{
					if(DOMUtils._re_tag.test(part))
						compiled = { descend : descend, type : "tag", data : part };
					else
						return null;
				}
			}
			if(compiled)
			{
				compiled.source = source;
				compiledselector.push(compiled);
				compiled = null;
				descend = "self";
			}
			if(nextselector)
			{
				if(!compiledselector.length)
					return null;
				compiledlist.push(compiledselector);
				compiledselector = [ ];
				nextselector = false;
				descend = "all";
			}
		}
		compiledlist.push(compiledselector);
		return compiledlist;
	},
	// Get sibling matches.
	_getSelectorSiblingMatch : function(parent, siblinglist)
	{
		var elementlist = [ ];
		for(var loop=0; loop < siblinglist.length; loop++)
		{
			var element = siblinglist[loop];
			var previousSibling = DOMUtils.getPreviousSiblingElement(element);
			if(element != parent && parent == previousSibling)
				elementlist.push(element);
		}
		return elementlist;
	},
	// Regular expression for parsing selector with attributes.
	_re_attrib 	: /^([A-Za-z]+)(|([~\^\*\$\|]?=)\"(.*)\")$/,
		// /^																Start
		// ([A-Za-z]+)				href								The attribute
		// (|																	Empty or...
		//		([~\^\*\$\|]?=)		^=									The comparison type ( ~= ^= *= $= |= = )
		//		\"(.*)\"					"http://"						Quoted string
		// )$/																End
	// Regular expression for matching tag names.
	_re_tag		: /^[A-Za-z]+$/
});
