/*
	utils.js
	Copyright (C) 2005 H. Adam Mill II
	
	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at
	
		 http://www.apache.org/licenses/LICENSE-2.0
	
	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
*/


// **** General ****

// Extend an object with another objects properties.
var extend = function(target, source)
{
	for(var property in source)
		target[property] = source[property];
};


// **** Array Utilities ****
var ArrayUtils =
{
	// Does the array contain the value.
	contains : function(list, value)
	{
		if(null == list)
			return false;	
		for(var loop=0; loop < list.length; loop++)
			if(value == list[loop])
				return true;
		return false;
	},
	// Convert an array-like object to an array.
	toArray : function(list)
	{
		if(null == list)
			return null;	
		else if(list instanceof Array)
			return list;
		var ary = [ ];
		for(var loop=0; loop < list.length; loop++)
			ary.push(list[loop]);
		return ary;
	},
	// Remove duplicate entries from an array.
	removeDuplicates : function(list)
	{
		if(null == list)
			return null;	
		var ary = [ ];
		for(var loop=0; loop < list.length; loop++)
		{
			if(!ArrayUtils.contains(ary, list[loop]))
				ary.push(list[loop]);
		}
		return ary;
	}	
};


// **** String Utilities ****
var StringUtils = 
{
	// Trim whitespace from both ends of a string.
	trim : function(value)
	{
		return value.replace(/(^\s+|\s+$)/g, "");
	},
	// Like String.split, only the seperator can be any of a group of characters 
	//  and the seperators are included in the array as individual items.
	breakup : function(value, chars)
	{
		var split = [ ];
		var current = "";
		for(var loop=0; loop < value.length; loop++)
		{
			var chr = value.substr(loop, 1);
			if((-1) == chars.indexOf(chr))
				current += chr;
			else
			{
				if("" != current)
					split.push(current);
				current = "";
				split.push(chr);
			}
		}
		if("" != current)
			split.push(current);
		return split;
	},
	// Count the number of times a string occurs within another string.
	count : function(value, search)
	{
		var count = 0;
		var idx = value.indexOf(search);
		while((-1) != idx)
		{
			count++;
			value = value.substr(idx+1);
			idx = value.indexOf(search);
		}
		return count;
	}
};

// **** RegExp Utilities ****
var RegExpUtils = 
{
	// Get a regular expression that finds a search item in a whitespace delimited list.
	getWhiteSpaceListMatcher : function(search)
	{
		return new RegExp("(^|\\s+)" + RegExpUtils.escape(search) + "($|\\s+)", "");
	},
	// Escape regexp control characters in a string.
	escape : function(value)
	{
		// Matches: . ? + * | ! \ / [ ] ( ) { } ^ $
		return value.replace(/([\.\?\+\*\|\!\\\/\[\]\(\)\{\}\^\$])/g,
			function(str)
			{
				return "\\" + str;
			});	
	}
};

var DOMUtils = 
{
	// Note Types
	ELEMENT_NODE : 1,
	// Get the previous sibling element (skipping over text and comment nodes).
	getPreviousSiblingElement : function(parent)
	{
		parent = DOMUtils.getElement(parent);
		var element = parent;
		do
		{
			if(!element.previousSibling)
				return null;
			element = element.previousSibling;
		}
		while(DOMUtils.ELEMENT_NODE != element.nodeType)
		return element;
	},
	// Get a child element by id.
	getElementById : function(parent, id, directchildren)
	{
		parent = DOMUtils.getElement(parent);
		if(parent.getElementById)
			return parent.getElementById(id);
		else
		{
			var itemlist = [ ];
			var children = parent.childNodes;
			if(children && 0 != children.length)
			{
				for (var loop=0; loop < children.length; loop++)
				{
					if(id == children[loop].id)
						return children[loop];
					if(!directchildren)
						if(children[loop].childNodes && 0 != children[loop].childNodes.length)
						{
							var element = DOMUtils.getElementById(children[loop], id);
							if(null != element)
								return element;
						}
				}
			}
		}
		return null;
	},
	// Get a direct child element by id.
	getChildById : function(parent, id)
	{
		return DOMUtils.getElementById(parent, id, true);
	},
	// Get elements by class name.
	getElementsByClassName : function(parent, className)
	{
		var re_classname = RegExpUtils.getWhiteSpaceListMatcher(className);
		var elementlist = DOMUtils.getElementsByFilter(parent, function(element) {
				if(element.className && re_classname.test(element.className))
						return true;
				return false;
			});
		return elementlist;
	},
	// Get direct children by class name.
	getChildrenByClassName : function(parent, className)
	{
		var re_classname = RegExpUtils.getWhiteSpaceListMatcher(className);
		return DOMUtils.getChildrenByFilter(parent, function(element) {
				if(element.className)
				{
					if(re_classname.test(element.className))
						return true;
				}
				return false;
			});			
	},
	// Get elements by type (text or comment).
	getElementsByType : function(parent, typename)
	{
		var elements = DOMUtils.getElementsByFilter(parent, function(element) {
				if("#"+typename == element.nodeName)
					return true;
				return false;
			});	
		return elements;
	},
	// Get all child nodes.
	getAllChildren : function(parent)
	{
		parent = DOMUtils.getElement(parent);
		if(parent.all)
			return parent.all;		
		else if(parent.getElementsByTagName)
			return parent.getElementsByTagName("*");
		else
			return [ ];
	},
	// Get the direct children of an element.
	getDirectChildren : function(parent)
	{
		parent = DOMUtils.getElement(parent);
		if(parent.childNodes)
			return parent.childNodes;
		else
			return [ ];
	},
	// Get direct children by tag name.
	getChildrenByTagName : function(parent, nodeName)
	{
		return DOMUtils.getChildrenByFilter(parent, function(element) {
				if(element.nodeName && element.nodeName.toLowerCase() == nodeName)
					return true;
				return false;
			});
	},
	// Get elements that pass a filter function.
	getElementsByFilter : function(parent, filter, extra, directchildren)
	{
		parent = DOMUtils.getElement(parent);
		var itemlist = [ ];
		var children = parent.childNodes;
		if(children && 0 != children.length)
		{
			for (var loop=0; loop < children.length; loop++)
			{
				if(filter(children[loop], extra))
					itemlist.push(children[loop]);
				if(!directchildren)
					if(children[loop].childNodes && 0 != children[loop].childNodes.length)
						itemlist = itemlist.concat(DOMUtils.getElementsByFilter(children[loop], filter, extra));
			}
		}
		return itemlist;
	},
	// Get direct child elements that pass a filter function.
	getChildrenByFilter : function(parent, filter, extra)
	{
		return DOMUtils.getElementsByFilter(parent, filter, extra, true);
	},	
	// Get an element.
	getElement : function(element)
	{
		if("string" == typeof element)
			return document.getElementById(element);
		else
			return element;	
	}
};
